#include <PalmOS.h>

#define FLASHBASE     (0x10c00000)
#define FLASH(a)      (*(((volatile unsigned short*)FLASHBASE)+a))

#define SOURCE_BASE			(*(void **)0x000054fc)
#define SOURCE_SIZE			(*(long  *)0x000054f8) // in bytes
#define TARGET_OFFSET		(*(long  *)0x000054f4) // in bytes
#define CKSUM				(*(long  *)0x000054f0)

/* erase sector */
int flash_sector_erase(long addr) {
  unsigned short status;

  /* initiate sector erase */
  FLASH(0x555)=0xaa;
  FLASH(0x2aa)=0x55;
  FLASH(0x555)=0x80;
  FLASH(0x555)=0xaa;
  FLASH(0x2aa)=0x55;

  /* ok, erase ... */
  FLASH(addr)=0x30;

  do {
    /* read status */
    status=FLASH(addr);

    /* test poll bit */
    if((status&0x80)==0x80) {
      return 0;   /* successfully erased! */
    }

    /* loop until timeout */
  } while((status&0x20)!=0x20);
  
  /* read status */
  status=FLASH(addr);
  
  if((status&0x80)==0x80)
    return 0;   /* successfully erased! */
  
  return 1;
}

/* write a single word to the flash rom */
int flash_write_word(unsigned short data, long addr) {
  unsigned short status;

  /* look if word already there */
  if(FLASH(addr)==data)
    return 1;

  /* initiate write */
  FLASH(0x555)=0xaa;
  FLASH(0x2aa)=0x55;
  FLASH(0x555)=0xa0;

  /* ok, write ... */
  FLASH(addr)=data;

  do {
    /* read status */
    status=FLASH(addr);

    /* test poll bit */
    if((status&0x80)==(data&0x80))
      return 1;   /* successfully written!!! */

    /* loop until timeout */
  } while((status&0x20)!=0x20);
  
  /* read status */
  status=FLASH(addr);

  if((status&0x80)==(data&0x80))
    return 1;   /* successfully written!!! */
  
  return 0;  /* operation failed :-( */
}

/* write data to flash rom (this is the dangerous part) */
/* this returns 0 on success, otherwise the address that */
/* caused the error */
long  flash_write(unsigned short *src, long dest, long size) {
  long i;
  unsigned short dummy;

  for(i=0;i<size;i++) {
    /* write as long as everything is fine */
    if(!flash_write_word(*src++, dest+i)) {
      /* oops, error -> return address */
      return(dest+i);
    }
  }

  return 0;  /* everything is fine */
}

void cksum (void)
{
	long ck = 0;
	int i;
	long * data = (long *)SOURCE_BASE;
	int size = SOURCE_SIZE / sizeof (long);

	for (i = 0; i < size; i++)
		ck += data[i];

	CKSUM = ck;
}

void flash_rom (void)
{
	flash_sector_erase (TARGET_OFFSET / 2);
	flash_write (SOURCE_BASE, TARGET_OFFSET / 2, SOURCE_SIZE / 2);
}

UInt32 PilotMain(UInt16 cmd, void *cmdBPB, UInt16 launchFlags)
{
        return 0;
}

